import java.awt.*;
import java.awt.event.*;
import java.io.* ;
import java.lang.Thread ;
import java.util.* ;
import java.net.*;


  // class that does the bulk of the work

class SchedManager extends Canvas implements Runnable, MouseListener
{
  BART applet ;
  Clock schedClock;
  Thread trainThread = null ;		// trains thread
  Vector activeVehicles = new Vector();
  TransitInfo transitInfo[];
  Image img ;				// train route map
  Image stopImg;
  Image stopInImg;
  Image playImg;
  Image playInImg;
  boolean scheduleLoaded=false;	//  schedule loaded and data initialized
  boolean end_BART;			//  schedule finished or not
  boolean imagesLoaded=false;
  boolean clockChanged=true;
  boolean trainsPaused=true;		// trains paused or not
  boolean pausedStatusChanged=true;	// trains just paused or resumed
  boolean firstDraw=true;		//  first repaint
  boolean inPaint ;

  Bline blines[] ;		//  line data structure
  int n_blines ;		//  number of lines
  int lines_active ;		//  number of lines currently active
  int start_order[] ;		//  order in which lines are started
  int betwLinesMs[] ;	        //  time in ms between starting times of lines
  int offsetMs=0;
  int nextRun[] ;
  int width;
  int height;

  SchedManager(BART app)
  {
    applet = app ;
    getImages();
    width = playImg.getWidth(this);
    height = playImg.getHeight(this);
    addMouseListener(this);
  }
    

  void getImages()
  {
    URL url=null;
	// route map
    try { url = new URL(applet.getCodeBase().toString()+"bart.gif") ; }
    catch(MalformedURLException e) {}
    img = applet.getImage(url) ;
	// stop button popped out
    try { url = new URL(applet.getCodeBase().toString()+"stop.gif") ; }
    catch(MalformedURLException e) {}
    stopImg = applet.getImage(url) ;
	// stop button pressed in
    try { url = new URL(applet.getCodeBase().toString()+"stop.in.gif") ; }
    catch(MalformedURLException e) {}
    stopInImg = applet.getImage(url) ;
	// play button popped out
    try { url = new URL(applet.getCodeBase().toString()+"play.gif") ; }
    catch(MalformedURLException e) {}
    playImg = applet.getImage(url) ;
	// play button pressed in
    try { url = new URL(applet.getCodeBase().toString()+"play.in.gif") ; }
    catch(MalformedURLException e) {}
    playInImg = applet.getImage(url) ;
    applet.do_wait(img) ;
    applet.do_wait(stopImg) ;
    applet.do_wait(stopInImg) ;
    applet.do_wait(playInImg) ;
    applet.do_wait(playImg) ;
    imagesLoaded = true;
  }


    //  does all the graphics
  synchronized void paint_trains(Graphics g, boolean win_exposed)
  {
    inPaint = true ;
      //  if schedule has been read, trains initialized, etc.
    if (scheduleLoaded)
    {
        // if expose event, or first repaint, redraw static part of image
      if (win_exposed || firstDraw)
      {
        repaint_backgr(g) ;
        firstDraw = false ;
      }
      else if (pausedStatusChanged)
      {
        if (!trainsPaused)
        {
          g.drawImage(stopImg, 339, 53, this);
          g.drawImage(playInImg, 430, 53, this);
        }
        else
        {
          g.drawImage(stopInImg, 339, 53, this);
          g.drawImage(playImg, 430, 53, this);
        }
      }
      if (!trainsPaused)
      {
        TransitVehicle currentVehicle;

          //  redraw all active vehicles
        for (Enumeration e = activeVehicles.elements();
             e.hasMoreElements();)
        {
          currentVehicle = (TransitVehicle) e.nextElement();
          currentVehicle.paint(g, img, this, applet.train_size);
        }
          //  redraw number-of-trains-running info if necessary
        for (int i = 0; i < n_blines/2 ; i++)
          transitInfo[i].paint(g, false);
      }
      schedClock.paint(g, false, end_BART);
      win_exposed = false ;
    }
    else if (imagesLoaded)
    {
      g.drawImage(stopInImg, 339, 53, this);
      g.drawImage(playImg, 430, 53, this);
      g.drawImage(img, 0, 0, this) ;
      g.setColor(Color.red);
      g.drawString("INITIALIZING",377,40) ;
    }
    inPaint = false ;
      //  paint done, train info can be updated again
    notifyAll() ;
  }


    // called by expose
  public void paint(Graphics g)
  {
    paint_trains(g,true) ;
  }


    // called by repaint
  public void update(Graphics g)
  {
    paint_trains(g,false);
  }


    // repaint static part of display
  void repaint_backgr(Graphics g)
  {
    g.drawImage(img, 0, 0, this) ;
    if (!trainsPaused)
    {
      g.drawImage(stopImg, 339, 53, this);
      g.drawImage(playInImg, 430, 53, this);
    }
    else
    {
      g.drawImage(stopInImg, 339, 53, this);
      g.drawImage(playImg, 430, 53, this);
    }
    g.setColor(Color.gray) ;
      // outline train legend area
    g.drawRect(7,6,190,169) ;
      // outline BART legend area
    g.drawRect(561,204,183,181) ;
      // scale bar
    g.fillRect(41,145,110,4) ;
    g.setColor(Color.blue) ;
    g.drawString("BART SIMULATION - LBNL",356,20) ;
    g.drawString("SF",143,379) ;
    g.setColor(Color.black) ;
    g.drawString("TRAINS",13,20) ;
    g.drawString("LINE",105,20) ;
    g.drawString("SAN FRANCISCO BAY",188,467) ;
      // draw scale text
    g.drawString("5 miles",72,140) ;
    g.drawString("8 km",72,165) ;
    if (!trainsPaused)
      schedClock.paint(g, true, end_BART);
      //  redraw number-of-trains-running info
    for (int i = 0; i < n_blines/2 ; i++)
      transitInfo[i].paint(g, true);
  }

  public void mouseEntered (MouseEvent evt){}
  public void mouseExited (MouseEvent evt){}
  public void mouseReleased (MouseEvent evt){}
  public void mouseClicked(MouseEvent evt){}

  public void mousePressed (MouseEvent evt)
  {
    int x, y;

    x = evt.getX();
    y = evt.getY();
    if ((y >= 53) &&
        (y < (53 + height)))
    {
        //  if stop "button" pushed
      if ((x >= 339) &&
          (x < (339 + width)))
      {
        System.gc();
        pauseThreads(true);
      }
        //  play "button" pushed
      else if ((x >= 430) &&
               (x < (430 + width)))
      {
        pauseThreads(false);
      }
    }
    return;
  }

/****   end of graphics section and start of schedule running section ****/

 
    //  wait for repaint to finish
  synchronized void sync_paint()
  {
    while (inPaint == true)
    {
      try { wait() ; }
      catch(InterruptedException e) {}
    }
  }


      // start trains thread
  void start_it()
  {
    if (trainThread == null)
    {
      trainThread = new Thread(this,"-3") ;
      trainThread.start() ;
    }
  }


    // (re)initialize train data
  void init_trains()
  {
    inPaint = true ;
    repaint() ;
      // force repaint
    sync_paint() ;
    end_BART = false ;	//  schedule not finished
    clockChanged=true;
    if (!scheduleLoaded)
    {
    pausedStatusChanged=true;	// trains just paused or resumed
    TransitInfo.setColors();
    SchedReader reader = new SchedReader(applet);
    n_blines = reader.n_blines;
    blines = reader.blines;
      //  same information in both directions
    transitInfo = new TransitInfo[n_blines/2];
    for (int i = 0 ; i < n_blines/2 ; i++)
      transitInfo[i] = new TransitInfo(i);
    start_order = reader.start_order;
    scheduleLoaded = true;
    }
    else
    {
      trainsPaused=false;		// trains paused or not
      pausedStatusChanged=true;	// trains just paused or resumed
    }
      //  suggest garbage collection
    System.gc() ;
  }



    // schedule-running thread
  public void run()
  {
    int i ;

      // initialize global data structures
    init_trains() ;
    nextRun = new int[n_blines];
      // run the schedule
    runSched() ;
    trainThread = null;
  }


        // runs the trains
  void runSched()
  {
    int i, j, k ;
    int curr_line ;		//  current line
    int diff_ms ;		//  difference in ms between last update and this
    int sleep_time;
    int prev_timefac = 1;
    TransitVehicle currentVehicle;
    TransitVehicle newVehicle;

    betwLinesMs = new int[n_blines] ;
      // do train thread ;
    while (!end_BART)
    {
      if (clockChanged)
      {
          //  first line starts immediately
        betwLinesMs[start_order[0]] = 0 ;
          //  waits between starting time of first run of one line and
          //  starting time of first run of another
        for (i = 1 ; i < n_blines ; i++)
        {
          betwLinesMs[start_order[i]] = (blines[start_order[i]].start_ms -
		    blines[start_order[i-1]].start_ms) ;
        }
        for (i = 0 ; i < n_blines ; i++)
        {
          curr_line = start_order[i];
          for (k = 0 ; k < blines[curr_line].nruns; k++)
          {
            blines[curr_line].workingStDiffs[k] =
              blines[curr_line].startDiffsMs[k];
          }
        }
        lines_active = 0;
        if (schedClock == null)
          schedClock = new Clock(applet.timefac, applet.timefac,
                               blines[start_order[0]].start_ms + offsetMs -
                               Clock.STATION_SLEEP);
        else
          schedClock.update(true, blines[start_order[0]].start_ms + offsetMs);
        if (offsetMs > 0)
        {
          findPositions();
          offsetMs = 0;
        }
        clockChanged = false;
        firstDraw = true;
        prev_timefac = schedClock.getRealMode();
      }
      else if (prev_timefac != schedClock.getRealMode())
      {
        schedClock.update(false, 0);
        prev_timefac = schedClock.getRealMode();
      }
      else 
          //  if any train positions updated, this will be set to false
        end_BART = true ;
      trainsPaused = false;
      diff_ms = schedClock.timeElapsed();
        // see if time to start new line
      for (i = lines_active; i < n_blines; i++)
      {
        curr_line = start_order[i] ;
          //  decrement wait time between lines
        betwLinesMs[curr_line] -= diff_ms ;
        if (betwLinesMs[curr_line] <= 0)
        {
          newVehicle = blines[curr_line].transitVehicles[0];
          newVehicle.reset(Clock.STATION_SLEEP+
                          betwLinesMs[curr_line]);
          activeVehicles.addElement(newVehicle);
              // update numbers-of-trains-active info
          transitInfo[(newVehicle.lineid)/2].incr_train_num(1);
          nextRun[curr_line] = 1 ;
          lines_active++ ;
          if ((lines_active < n_blines) &&
            (blines[curr_line].start_ms != blines[start_order[lines_active]].start_ms))
            break;
        }
        else
          break;
      }
        // see if time to start new run
      for (i = 0 ; i < lines_active ; i++)
      {
          //  current line according to starting order
        curr_line = start_order[i] ;
        for (k = nextRun[curr_line]; k < blines[curr_line].nruns; k++)
        {
          if (blines[curr_line].startDiffsMs[k] != 0.0)
          {
              //  decrement wait between runs time
            blines[curr_line].workingStDiffs[k] -= diff_ms ;
          }
            //  if time to start
          if (blines[curr_line].workingStDiffs[k] <= 0)
          {
            newVehicle = blines[curr_line].transitVehicles[k];
            newVehicle.reset(Clock.STATION_SLEEP +
                          blines[curr_line].workingStDiffs[k]);
            activeVehicles.addElement(newVehicle);
              // update numbers-of-trains-active info
            transitInfo[(newVehicle.lineid)/2].incr_train_num(1);
            nextRun[curr_line]++ ;
          }
          else
            break;
        }
      }
        // update all vehicle positions
      for (Enumeration e = activeVehicles.elements();
           e.hasMoreElements();)
      {
        currentVehicle = (TransitVehicle) e.nextElement();
        if (currentVehicle.beenRemoved())
        {
          transitInfo[(currentVehicle.lineid)/2].incr_train_num(-1);
          activeVehicles.removeElement(currentVehicle);
        }
        else
        {
          end_BART = false;
            // update vehicle
          currentVehicle.update(diff_ms, applet.train_size, false);
        }
      }
      inPaint = true ;
      repaint() ;
        // force repaint
      sync_paint() ;
      schedClock.adjustRate(diff_ms);
      if (end_BART)
      {
        for (i = 0; i < n_blines/2 ; i++)
          transitInfo[i].reset_train_num();
      }
      if (schedClock.getRealMode() == 1)
      {
        sleep_time = 500;
      }
      else
      {
        sleep_time = 15;	// runs fast enough, and much smoother
      }
      try { trainThread.sleep(sleep_time) ;}
      catch(InterruptedException e) {}
    }
  }


  void findPositions()
  {
    int i, j, k ;
    int curr_line ;		//  current line
    boolean remove_line ;
    int sumRunDiffs;
    int sumStationDiffs;
    TransitVehicle newVehicle;
    int diffMs=0;
    int endingTime;
    int lineStartMs[] = new int[n_blines];

    for (i = 0; i < n_blines; i++)
    {
      curr_line = start_order[i] ;
      lineStartMs[curr_line] = blines[start_order[i]].start_ms -
                              (blines[start_order[0]].start_ms + offsetMs) -
                              Clock.STATION_SLEEP;
      nextRun[curr_line] = 0;
      if (lineStartMs[curr_line] <= 0)
      {
        betwLinesMs[curr_line] = 0;
        lines_active++ ;
      }
      else
      {
        betwLinesMs[curr_line] -= lineStartMs[curr_line];
        break;
      }
    }

    for (i = 0 ; i < lines_active ; i++)
    {
      remove_line = true;
        //  current line according to starting order
      curr_line = start_order[i] ;
      sumRunDiffs = 0;
      k = 0;
        // find all runs that begin before the new time
      while ((sumRunDiffs <= (-lineStartMs[curr_line])) &&
             (k < blines[curr_line].nruns))
      {
        endingTime = (sumRunDiffs +
          blines[curr_line].transitVehicles[k].totalTime) + lineStartMs[curr_line];
            //  start all runs that don't end before new time
        if (endingTime > 0)
        {
          remove_line = false;
          end_BART = false;
          newVehicle = blines[curr_line].transitVehicles[k];
          newVehicle.reset(Clock.STATION_SLEEP);
          activeVehicles.addElement(newVehicle);
          diffMs = -lineStartMs[curr_line] - sumRunDiffs;
            // update vehicle
          newVehicle.update(diffMs,applet.train_size,true);
            // update numbers-of-trains-active info
          transitInfo[(newVehicle.lineid)/2].incr_train_num(1);
        }
        nextRun[curr_line]++;
        if ((k+1) < blines[curr_line].nruns)
          sumRunDiffs += blines[curr_line].workingStDiffs[k+1];
        k++;
      }
        //  kludge
      if (remove_line)
        nextRun[curr_line] = 10000;
      else if (nextRun[curr_line] < blines[curr_line].nruns)
      {
        blines[curr_line].workingStDiffs[nextRun[curr_line]] = (sumRunDiffs) + lineStartMs[curr_line];
      }
    }
  }


    // stop the train thread
  void stop_threads()
  {
    if (trainsPaused)
      trainThread.resume() ;
    trainsPaused = true ;
    activeVehicles.removeAllElements();
    trainThread.stop() ;
    trainThread = null ;
  }


    //  pause or continue trains based on button click
  void pauseThreads(boolean pause_it)
  {
      // if not already paused, pause it
    if (!trainsPaused && pause_it)
    {
      trainThread.suspend() ;
      schedClock.suspendTime();
      trainsPaused = true ;
      pausedStatusChanged = true;
    }
      // if not already running, resume thread
    else if (trainsPaused && !pause_it)
    {
      schedClock.reStartTime() ;
      trainThread.resume() ;
      trainsPaused = false ;
      pausedStatusChanged = true;
    }
  }


    //  real-time or speeded-up mode	
  void real_mode(boolean real_choice)
  {
    schedClock.setRealMode(real_choice);
  }


  boolean changeTime(String timeStr)
  {
    int currms = schedClock.getDailyTimeinMS(timeStr);

    if (currms == -1)
      return false;
    else
    {
      offsetMs = currms - blines[start_order[0]].start_ms;
      if (offsetMs < 0)
        return false;
      clockChanged = true;
      activeVehicles.removeAllElements();
      for (int i = 0; i < n_blines/2 ; i++)
        transitInfo[i].reset_train_num();
      return true;
    } 
  }
}
